library ieee;
use ieee.std_logic_1164.all;

entity MercurioIV_RV16Cm is
	port
	(
		CLOCK_50MHz	: in bit;
		
		KEY			: in bit_vector(11 downto 0);
		
		SA, SB		: in bit_vector(7 downto 0);
		SW				: in bit_vector(3 downto 0);

		LED_R 		: out bit;
		LED_G 		: out bit;
		LCD_EN		: out bit;
		LCD_RW		: out bit;
		LCD_RS		: out bit;
		LCD_D			: inout bit_vector(7 downto 0);
		
		DISP0_D		: out bit_vector(7 downto 0);
		DISP1_D		: out bit_vector(7 downto 0);
		
      VGA_R 		: OUT bit_vector(3 downto 0);
		VGA_G 		: OUT bit_vector(3 downto 0);
		VGA_B 		: OUT bit_vector(3 downto 0);
      VGA_HS 		: OUT bit;
		VGA_VS 		: OUT bit
	);

end MercurioIV_RV16Cm;


architecture top of MercurioIV_RV16Cm is

	signal clk			: bit;
	signal reset		: bit;
	signal clkn			: bit;
	signal clk_cnt		: bit_vector(25 downto 0);
	
	signal num_state	: bit_vector(3 downto 0);

	constant ZeroCnst	: bit_vector(15 downto 0) := (others => '0');
	constant DoisCnst	: bit_vector(15 downto 0) := (1 => '1', others => '0');

	--Internal signals
	signal PC			: bit_vector(15 downto 0);
	--signal PCNext		: bit_vector(15 downto 0);
	signal Adr			: bit_vector(15 downto 0);
	signal Instr		: bit_vector(15 downto 0);
	signal ReadData	: bit_vector(15 downto 0);
	signal Data			: bit_vector(15 downto 0);
	signal ImmExt		: bit_vector(15 downto 0);
	signal RD1			: bit_vector(15 downto 0);
	signal RD2			: bit_vector(15 downto 0);
	signal WriteData	: bit_vector(15 downto 0);
	signal B				: bit_vector(15 downto 0);
	signal SrcA			: bit_vector(15 downto 0);
	signal SrcB			: bit_vector(15 downto 0);
	signal ALUResult	: bit_vector(15 downto 0);
	signal ALUOut		: bit_vector(15 downto 0);
	signal Result		: bit_vector(15 downto 0);
	signal Zero			: bit;
--	signal Op			: bit_vector(1 downto 0);
--	signal Funct3		: bit_vector(2 downto 0);
	signal PCWrite		: bit;
	signal AdrSrc		: bit;
	signal MemWrite	: bit;
	signal IRWrite		: bit;
	signal RegWrite	: bit;
	signal ResultSrc	: bit_vector(1 downto 0);
	signal ALUControl	: bit_vector(2 downto 0);
	signal ALUSrcA		: bit_vector(1 downto 0);
	signal ALUSrcB		: bit_vector(1 downto 0);
	signal ImmSrc		: bit_vector(1 downto 0);
	
	alias Op: bit_vector(1 downto 0) is Instr(1 downto 0);
	alias Funct3: bit_vector(2 downto 0) is Instr(15 downto 13);
	alias Rs1: bit_vector(2 downto 0) is Instr(7 downto 5);
	alias Rs2: bit_vector(2 downto 0) is Instr(4 downto 2);
	alias Rd: bit_vector(2 downto 0) is Instr(7 downto 5);
	alias Imm8U: bit_vector(4 downto 0) is Instr(12 downto 8);
	alias Imm8D: bit_vector(2 downto 0) is Instr(4 downto 2);
--	alias Addr11: bit_vector(10 downto 0) is Instr(12 downto 2);
	alias Addr11U: bit_vector(2 downto 0) is Instr(7 downto 5);
	
	signal ReadData_std: std_logic_vector(15 downto 0);
	
begin	
	--Unit Control
	FSM0: work.FSM(estrutural)
		port map
		(
			--Entradas
			clk => clk, 
			reset => reset,
			Zero => Zero,
			Op => Op,
			Funct3 => Funct3,
			-- Saidas
			PCWrite => PCWrite,
			AdrSrc => AdrSrc,
			MemWrite => MemWrite,
			IRWrite => IRWrite,
			RegWrite => RegWrite,
			ResultSrc => ResultSrc,
			ALUControl => ALUControl,
			ALUSrcA => ALUSrcA,
			ALUSrcB => ALUSrcB,
			ImmSrc => ImmSrc,
			num_state => num_state,
			ZFlag_debug => LED_G
		);

	--PC Register
	PC0: process (clk, reset)
	begin
		if (reset = '1') then
			PC <= (others => '0');
		elsif (clk'event and clk='1') then
			if (PCWrite='1') then
				PC <= Result;
			end if;
		end if;
	end process;
	
	--Memory Address source selection
	AdrSrc0: work.mux21
		generic map
		(
			n => 16
		)
		port map
		(
			I1 => Result,
			I0 => PC,
			S => AdrSrc,
			Z => Adr
		);
	
	--Program/Data Memory
	Memory0 : work.Memory
		port map
		(
			address	 => to_stdlogicvector('0' & Adr(10 downto 1)),
			clock	 => to_stdUlogic(clkn),
			data	 => to_stdlogicvector(WriteData),
			wren	 => to_stdUlogic(MemWrite),
			q	 => ReadData_std
		);
	ReadData <= to_bitvector(ReadData_std);

	--IR Register
	IR0: process (clk)
	begin
		if (clk'event and clk='1') then
			if (IRWrite='1') then
				Instr <= ReadData;
			end if;
		end if;
	end process;

	--Data Register
	DR0: process (clk)
	begin
		if (clk'event and clk='1') then
				Data <= ReadData;
		end if;
	end process;
	
	--Register Bank
	RegBank0: work.RegBank--(estrutural)
		generic map
		(
			WIDTH => 16,
			SIZE => 3
		)
		port map
		(
			--Entradas
			CLK => clk,
			WE3 => RegWrite,
			A1 => Rs1,
			A2 => Rs2,
			A3 => Rd,
			WD3 => Result,
			
			-- Saidas
			RD1 => RD1,
			RD2 => RD2
		);

	--Imm source selection
	ImmSrc0: work.mux31
		generic map
		(
			n => 16
		)
		port map
		(
			I2 => "00000" & Addr11U & Imm8U & Imm8D,			--CBm-type					(op=11)
			I1 => (7 downto 0 => Imm8U(4)) & Imm8U & Imm8D,	--CSm-type | CLm-type	(op=10)
			I0 => (7 downto 0 => Imm8U(4)) & Imm8U & Imm8D,	--CIm-type					(op=01)
			S => ImmSrc,
			Z => ImmExt
		);
	
	--Reg Data
	Reg0: process (clk)
	begin
		if (clk'event and clk='1') then
				WriteData <= RD1;
				B <= RD2;
		end if;
	end process;

	--ALU A source selection
	ALUSrcA0: work.mux31
		generic map
		(
			n => 16
		)
		port map
		(
			I2 => WriteData,
			I1 => ZeroCnst,
			I0 => PC,
			S => ALUSrcA,
			Z => SrcA
		);

	--ALU B source selection
	ALUSrcB0: work.mux31
		generic map
		(
			n => 16
		)
		port map
		(
			I2 => DoisCnst,
			I1 => ImmExt,
			I0 => B,
			S => ALUSrcB,
			Z => SrcB
		);
	
	ALU0: work.ALU--(estrutural)
		generic map
		(
			WIDTH => 16
		)
		port map
		(
			--Entradas
			A => SrcA,
			B => SrcB,
			ALUControl => ALUControl,
			--Saidas
			Result => ALUResult,
			Zero => Zero
		);
	
	--ALUOut
	ALUOut0:process (clk)
	begin
		if (clk'event and clk='1') then
				ALUOut <= ALUResult;
		end if;
	end process;

	--ALU Ouput source selection
	ResultSrc0: work.mux31
		generic map
		(
			n => 16
		)
		port map
		(
			I2 => ALUResult,
			I1 => Data,
			I0 => ALUOut,
			S => ResultSrc,
			Z => Result
		);
	
	-- Component Instantiation Statement
	
	CLK0: work.counter(operadores)
		generic map
		(
			WIDTH => 26
		)
		port map
		(
			clk => CLOCK_50MHz,
			clr => '0',				--Free running
			q => clk_cnt
		);
	
	clk <= clk_cnt(25);
	clkn <= not clk;
	
	reset <= KEY(11);
	
	LCD0: work.LCD(rtl)
		port map
		(
			clk => CLOCK_50MHz, 
			
			Instr => Instr,
			IRWrite => IRWrite,
			PC => PC,
			PCWrite => PCWrite,

			LCD_EN => LCD_EN,
			LCD_RW => LCD_RW,
			LCD_RS => LCD_RS,
			LCD_D => LCD_D
		);
		
	display1: work.decod(operadores)
		port map (
				PC(3), PC(2), PC(1), PC(0),
				DISP1_D(0), DISP1_D(1), DISP1_D(2), DISP1_D(3), DISP1_D(4), DISP1_D(5), DISP1_D(6)--, --DISP1_D(7)
				);

	display0: work.decod(operadores)
		port map (
				num_state(3), num_state(2), num_state(1), num_state(0), 
				DISP0_D(0), DISP0_D(1), DISP0_D(2), DISP0_D(3), DISP0_D(4), DISP0_D(5), DISP0_D(6)--, --DISP0_D(7)
				);

	LED_R <= clk;
	DISP1_D(7) <= PCWrite;
	DISP0_D(7) <= IRWrite;
	
	VGA0: work.VGA_Generator
		port map (
			CLOCK_50MHz => CLOCK_50MHz,
			KEY => KEY,
			SW => SW,
			VGA_R => VGA_R,
			VGA_G => VGA_G,
			VGA_B => VGA_B,
			VGA_HS => VGA_HS,
			VGA_VS => VGA_VS,
			--Debug signals
			clk => clk,
			reset => reset,
			PC => PC,
			Adr => Adr,
			Instr => Instr,
			ReadData => ReadData,
			RD1 => RD1,
			RD2 => RD2,
			ALUResult => ALUResult,
			Zero => Zero,
			PCWrite => PCWrite,
			RegWrite => RegWrite
		);

end top;
